package org.processmining.visualisation3d;

import java.awt.Graphics;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.text.DecimalFormat;
import java.util.Vector;

import javax.swing.JPanel;

import org.processmining.visualisation3d.garbagecollection.GLGarbageCollector;
import org.processmining.visualisation3d.gldatastructures.IGLResource;
import org.processmining.visualisation3d.graphicsdatastructures.JoglCamera;
import org.processmining.visualisation3d.measurements.AverageTimeMeasure;
import org.processmining.visualisation3d.scenemanagement.JoglScene;
import org.processmining.visualisation3d.settings.JoglGraphicsSettings;

import com.jogamp.opengl.GL2;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLEventListener;
import com.jogamp.opengl.awt.GLCanvas;


public class JoglPanel extends JPanel implements ComponentListener, GLEventListener
{

	/**
	 * GUI panel
	 * Encapsulates a GLCanvas object
	 */
	private static final long serialVersionUID = -6689258773462230598L;
	
	private GLCanvas canvas;
	private JoglCamera cam;
	private JoglScene scene;
	private GraphicsDisplay display = null;
	private JoglInteractionManager interactionManager;
	private boolean bContextInitialized = false;
	private boolean bContentLoaded = false;

	private JoglGraphicsSettings graphicsSettings = null;

	public Vector<JoglPreRenderTask> prerenderTasks = new Vector<JoglPreRenderTask>();
	
	private AverageTimeMeasure avgUpdateTime = new AverageTimeMeasure();
	//private AverageTimeMeasure avgRenderTime = new AverageTimeMeasure();
	
	public static GLGarbageCollector gc = new GLGarbageCollector();
	
	public JoglGraphicsSettings getGraphicsSettings()
	{
		return graphicsSettings;
	}
	
	public GraphicsDisplay getDisplay()
	{
		return display;
	}
	
	public JoglCamera getCamera()
	{
		return cam;
	}
	
	public JoglPanel()
	{
		CreateGLView();
	}

	private void CreateGLView() 
	{
		canvas = GraphicsDisplay.CreateGLCanvas();
		
		if (canvas != null)
		{
			System.out.println("Successfully created GLContext!");
			bContextInitialized = true;

			display = new GraphicsDisplay();
			interactionManager = new JoglInteractionManager(display);
			MouseInputArea mia = new MouseInputArea(this);
			mia.proc = interactionManager;
			//canvas.addGLEventListener(interactionManager);
			canvas.addGLEventListener(this);
			canvas.setSize(800, 600);
			add(canvas);
			
			canvas.addKeyListener(interactionManager);
			canvas.addMouseListener(mia);
			canvas.addMouseMotionListener(mia);
			canvas.addMouseWheelListener(mia);
			
			this.addComponentListener(this);
			
			interactionManager.InitAnimator(canvas);
			
			repaint();
		}
	}
	
	public boolean isContextInitialized()
	{
		return bContextInitialized;
	}
	
	public void setScene(JoglScene scene)
	{
		this.scene = scene;
		if (interactionManager != null)
		{
			interactionManager.setScene(scene);
		}
	}

    public void paintComponent(Graphics g) 
    {
    	canvas.display();
    }

    public void componentResized(ComponentEvent e) 
    {
        // Perform calculation here
    	System.out.println("Canvas size: " + this.getWidth() + "x" + this.getHeight());
    	canvas.setBounds(0, 0, this.getWidth(), this.getHeight());
    	canvas.display();

    	//not optimal, but request focus after the panel is actually displayed
		canvas.requestFocus();
    }

    public void componentHidden(ComponentEvent e) {}

    public void componentMoved(ComponentEvent e) {}

    public void componentShown(ComponentEvent e) {}
    
    public JoglInteractionManager getInteractionManager()
    {
    	return interactionManager;
    }
    
    public void setCamera(JoglCamera cam)
    {
    	this.cam = cam;
    }
    
    public void InitVR()
    {
    	interactionManager.InitVRAnimator(canvas);
    }
    
	@Override
	public void init(GLAutoDrawable glDrawable) 
	{
		System.out.println("Initializing GLCanvas...");
		final GL2 gl = glDrawable.getGL().getGL2();

		if (graphicsSettings == null)
		{
			graphicsSettings = JoglGraphicsSettings.GetDefaultSettings(gl);
			//graphicsSettings = JoglGraphicsSettings.GetLowSettings(gl);
		}
		
		JoglGraphicsSettings.CheckPlatformSupport(gl);
		
		GraphicsPipeline pipe = new GraphicsPipeline();
		pipe.gl = gl;
		//pipe.panel = this;
		pipe.gc = gc;
		pipe.settings = getGraphicsSettings();
		
		display.setGraphicsSettings(graphicsSettings);
		display.init(glDrawable, pipe);
		
		if (!bContentLoaded)
		{
			LoadContent(pipe);
			bContentLoaded = true;
		}

		gc.Reinit(pipe);
		
		interactionManager.getAnimator().start();
		interactionManager.getAnimator().pause();
	}
	
	//GLEventListener implementations
	@Override
	public void display(GLAutoDrawable glDrawable) 
	{
		final GL2 gl = glDrawable.getGL().getGL2();

        //System.out.println("display");
 
		if (scene != null)
		{
//			if (updateCounter == 0)
//			{
			try
			{
				avgUpdateTime.start();
				interactionManager.UpdateScene(scene);
				avgUpdateTime.stop();
			}
			catch (Exception ex)
			{
				System.out.println("Error in scene update!");
				ex.printStackTrace();
			}
//			}

//			updateCounter++;
//			if (updateCounter == 3)
//			{
//				updateCounter = 0;
//			}
			
			try
			{
				//avgRenderTime.start();
				GraphicsPipeline pipe = getGraphicsPipeline(gl);
				
				while (!prerenderTasks.isEmpty())
				{
					JoglPreRenderTask task = prerenderTasks.remove(0);
					task.Execute(pipe);
				}

				display.display(glDrawable, pipe);
				
				//avgRenderTime.stop();
			}
			catch (Exception ex)
			{
				System.out.println("Error in scene render!");
			}
		}

		GraphicsPipeline pipe = new GraphicsPipeline();
		pipe.gl = gl;
		pipe.gc = gc;
		pipe.settings = getGraphicsSettings();
		
		gc.DisposeGarbage(pipe);
		
		//MeasureFrameTime();
	}

	private GraphicsPipeline getGraphicsPipeline(final GL2 gl)
	{
		GraphicsPipeline pipe = new GraphicsPipeline();
		pipe.scene = scene;
		pipe.cam = cam;
		pipe.display = display;
		pipe.gl = gl;
		pipe.gc = gc;
		pipe.settings = getGraphicsSettings();
		return pipe;
	}
	
	@Override
	public void reshape(GLAutoDrawable glDrawable, int x, int y, int width, int height) 
	{
		display.reshape(glDrawable, x, y, width, height, getGraphicsPipeline(glDrawable.getGL().getGL2()));

//		if (scene != null && scene.cam != null)
//		{
//			scene.UpdateCameraLens(display.getTarget().getAspect());
//		}
		
		interactionManager.displaySizeChangedEvent(width, height);
	}
	
	@Override
	public void dispose(GLAutoDrawable gLDrawable) 
	{
		interactionManager.getAnimator().stop();
		
		GL2 gl = gLDrawable.getGL().getGL2();
		
		DecimalFormat df = new DecimalFormat("#.00");
		System.out.println("AvgFPS: " + interactionManager.getAnimator().getLastFPS());
		System.out.println("AvgUpdateTime: " + df.format(avgUpdateTime.getAverageMilliSeconds()) + "ms");
		System.out.println("AvgRenderTime: " + df.format(display.getAvgRenderTime().getAverageMilliSeconds()) + "ms");
		System.out.println("Disposing GLCanvas.");

		GraphicsPipeline pipe = new GraphicsPipeline();
		pipe.gl = gl;
		pipe.gc = gc;
		pipe.settings = getGraphicsSettings();
		
		gc.DisposeGarbage(pipe);
		gc.Uninit(pipe);
		
//		display.dispose(gLDrawable);
//		
//		
//		if (scene != null)
//		{
//			if (persist != null)
//			{
//				persist.Save(scene, cam, interactionManager, this);
//			}
//			scene.Dispose(gl);
//		}
//		
//		interactionManager.ClearInteractions();
//		
//		CleanupResources(gl);
		
		//scene = null;
	}

	public void LoadContent(GraphicsPipeline pipe) 
	{
//		if (defaultFont != null)
//		{
//			System.out.println("Loading default font.");
//			defaultFont.LoadFontTexture(gl, this);
//		}
		if (JoglResourceManager.HasResource("DefaultFont"))
		{
			System.out.println("Loading default font.");
			FontData defaultFont = (FontData)JoglResourceManager.GetResource("DefaultFont");
			defaultFont.LoadFontTexture(pipe);
		}
	}
	
	public void Close()
	{
		System.out.println("Closing JoglPanel");
		
		display.dispose(canvas);
//		
//		if (scene != null)
//		{
//			if (persist != null)
//			{
//				persist.Save(scene, cam, interactionManager, this);
//			}
//			scene.Dispose(gl);
//		}
//		
//		interactionManager.ClearInteractions();
//		
		CleanupResources();
	}
	
	public void CleanupResources() 
	{
		//cleaning some resources here, not the optimal place to do this but better than nothing
		for (String resName : JoglResourceManager.GetResourceNames())
		{
			Object res = JoglResourceManager.GetResource(resName);
			if (res instanceof IGLResource)
			{
				IGLResource t = (IGLResource)res;
				gc.QueueForDisposal(t);
			}
		}
		
//		if (defaultFont != null)
//		{
//			defaultFont.DisposeFontTexture(gl);
//			defaultFont = null;
//		}
//		if (JoglResourceManager.HasResource("DefaultFont"))
//		{
//			FontData defaultFont = (FontData)JoglResourceManager.GetResource("DefaultFont");
//			defaultFont.DisposeFontTexture(gl);
//		}
		JoglResourceManager.ClearResources();
		
		//GLGarbageCollector.DisposeGarbage(gl, true);
		gc.Finalize();
	}
	
//	private void MeasureFrameTime() 
//	{
//		avgSubmitTime.stop();
//		
//		if (count > 90)
//		{
//			double avgTimeMs = avgSubmitTime.getAverageMilliSeconds();
//			double avgFps = 1000.0 / avgTimeMs;
//			System.out.println("FPS: " + avgFps);
//			count = 0;
//		}
//		count++;
//		avgSubmitTime.start();
//	}
}
