package org.processmining.visualisation3d.gldatastructures;

import java.nio.IntBuffer;

import org.processmining.visualisation3d.GraphicsPipeline;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2;

public class JoglMultiSampleFrameBufferObject implements IJoglFrameBufferObject
{
	private boolean bLoaded = false;
	private int width = 1024;
	private int height = 1024;
	private int numSamples = 4;
	
	private int framebufferId = 0;
	private int status = 0;
	
	private IJoglFrameBufferAttachment depthbuffer = null;
	private IJoglFrameBufferAttachment colorbuffer = null;
	
	public static int GetMaxSamples(GL2 gl)
	{
		IntBuffer ib = IntBuffer.allocate(1);
		gl.glGetIntegerv(GL.GL_MAX_SAMPLES, ib);
		
		return ib.get(0);
	}
	
	private static int GenFramebuffer(GL2 gl)
	{
		IntBuffer ib = IntBuffer.allocate(1);
		gl.glGenFramebuffers(1, ib);
		
		return ib.get(0);
	}
	
	private static void DeleteFramebuffer(GL2 gl, int id)
	{
		IntBuffer ib = IntBuffer.allocate(1);
		ib.put(id);
		ib.position(0);
		gl.glDeleteFramebuffers(1, ib);
	}
	
	public int getSampleNum()
	{
		return numSamples;
	}
	
	public int getWidth()
	{
		return width;
	}
	
	public int getHeight()
	{
		return height;
	}
	
	public int getFramebufferId()
	{
		return framebufferId;
	}
	
	public int getColorbufferId()
	{
		JoglMultiSampleRenderBufferObject tmp = (JoglMultiSampleRenderBufferObject)colorbuffer;
		return tmp.getRenderBufferId();
	}
	
	public JoglMultiSampleRenderBufferObject getColorbuffer()
	{
		JoglMultiSampleRenderBufferObject tmp = (JoglMultiSampleRenderBufferObject)colorbuffer;
		return tmp;
	}
	
	public JoglMultiSampleFrameBufferObject(GraphicsPipeline pipe, int width, int height, int numSamples)
	{
		this.width = width;
		this.height = height;
		this.numSamples = numSamples;

		load(pipe);		
	}
	
	@Override
	public boolean isLoaded()
	{
		return bLoaded;
	}
	
	@Override
	public void load(GraphicsPipeline pipe)
	{
		GL2 gl = pipe.gl;
		framebufferId = GenFramebuffer(gl);

		Bind(pipe);
		if (depthbuffer != null)
		{
			if (!depthbuffer.isLoaded())
			{
				depthbuffer.load(pipe);
			}
			AttachDepthBuffer(pipe, depthbuffer);
		}
		if (colorbuffer != null)
		{
			if (!colorbuffer.isLoaded())
			{
				colorbuffer.load(pipe);
			}
			AttachColorBuffer0(pipe, colorbuffer);
		}
		Unbind(pipe);

		if (pipe.gc != null)
		{
			pipe.gc.Track(this);
		}
		bLoaded = true;
	}

	@Override
	public void unload(GraphicsPipeline pipe)
	{
		DeleteFramebuffer(pipe.gl, framebufferId);
		framebufferId = 0;

		if (pipe.gc != null)
		{
			pipe.gc.Untrack(this);
		}
		bLoaded = false;
	}
	
	public IJoglFrameBufferAttachment CreateRboAttachment(GraphicsPipeline pipe, int bufferType)
	{
		return new JoglMultiSampleRenderBufferObject(pipe, width, height, numSamples, bufferType);
	}
	
	public boolean isFramebufferComplete()
	{
		if (status != GL.GL_FRAMEBUFFER_COMPLETE)
		{
			return false;
		}
		return true;
	}
	
	private void AttachRenderBuffer(GraphicsPipeline pipe, int attachPoint, IJoglFrameBufferAttachment rbo)
	{
//		gl.glBindRenderbuffer(GL.GL_RENDERBUFFER, rbo.getRenderBufferId());
//		gl.glFramebufferRenderbuffer(GL.GL_FRAMEBUFFER, attachPoint, GL.GL_RENDERBUFFER, rbo.getRenderBufferId());
//		gl.glBindRenderbuffer(GL.GL_RENDERBUFFER, 0);	
		rbo.attach(pipe, attachPoint);
	}
	
	public void AttachDepthBuffer(GraphicsPipeline pipe, IJoglFrameBufferAttachment rbo)
	{
		AttachRenderBuffer(pipe, GL.GL_DEPTH_ATTACHMENT, rbo);
		depthbuffer = rbo;

		status = pipe.gl.glCheckFramebufferStatus(GL.GL_FRAMEBUFFER);
	}
	
	public void AttachColorBuffer0(GraphicsPipeline pipe, IJoglFrameBufferAttachment rbo)
	{
		AttachRenderBuffer(pipe, GL.GL_COLOR_ATTACHMENT0, rbo);
		colorbuffer = rbo;

		status = pipe.gl.glCheckFramebufferStatus(GL.GL_FRAMEBUFFER);
	}
	
	public void Dispose(GraphicsPipeline pipe)
	{
		unload(pipe);
		
		if (depthbuffer != null)
			depthbuffer.unload(pipe);
		if (colorbuffer != null)
			colorbuffer.unload(pipe);
	}
	
	@Override
	public void Bind(GraphicsPipeline pipe)
	{
		pipe.gl.glBindFramebuffer(GL.GL_FRAMEBUFFER, framebufferId);
	}
	
	@Override
	public void Bind(GraphicsPipeline pipe, int type)
	{
		pipe.gl.glBindFramebuffer(type, framebufferId);
	}
	
	@Override
	public void Unbind(GraphicsPipeline pipe)
	{
		pipe.gl.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0);
	}
	
	public void BlitToFramebuffer(GraphicsPipeline pipe, IJoglFrameBufferObject fbo)
	{
		GL2 gl = pipe.gl;
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, framebufferId);
		gl.glReadBuffer(GL.GL_COLOR_ATTACHMENT0);
		gl.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, fbo.getFramebufferId());
		gl.glDrawBuffer(GL.GL_COLOR_ATTACHMENT0);
		
		gl.glBlitFramebuffer(0, 0, width, height, 0, 0, fbo.getWidth(), fbo.getHeight(), GL.GL_COLOR_BUFFER_BIT, GL.GL_NEAREST);
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, 0);
		gl.glReadBuffer(GL.GL_FRONT);
	}
	
	public void BlitToFramebuffer(GraphicsPipeline pipe, IJoglFrameBufferObject fbo, int filter)
	{
		GL2 gl = pipe.gl;
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, framebufferId);
		gl.glReadBuffer(GL.GL_COLOR_ATTACHMENT0);
		gl.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, fbo.getFramebufferId());
		gl.glDrawBuffer(GL.GL_COLOR_ATTACHMENT0);
		
		gl.glBlitFramebuffer(0, 0, width, height, 0, 0, fbo.getWidth(), fbo.getHeight(), GL.GL_COLOR_BUFFER_BIT, filter);
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, 0);
		gl.glReadBuffer(GL.GL_FRONT);
	}
	
	public void BlitToBackbuffer(GraphicsPipeline pipe, int w, int h)
	{
		GL2 gl = pipe.gl;
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, framebufferId);
		gl.glReadBuffer(GL.GL_COLOR_ATTACHMENT0);
		gl.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, 0); //unbind any bound draw framebuffers
		gl.glDrawBuffer(GL.GL_BACK);
		
		gl.glBlitFramebuffer(0, 0, width, height, 0, 0, w, h, GL.GL_COLOR_BUFFER_BIT, GL.GL_NEAREST);
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, 0);
		gl.glReadBuffer(GL.GL_FRONT);
	}
	
	public void BlitToFrontbuffer(GraphicsPipeline pipe, int w, int h)
	{
		GL2 gl = pipe.gl;
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, framebufferId);
		gl.glReadBuffer(GL.GL_COLOR_ATTACHMENT0);
		gl.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, 0); //unbind any bound draw framebuffers
		gl.glDrawBuffer(GL.GL_FRONT);
		
		gl.glBlitFramebuffer(0, 0, width, height, 0, 0, w, h, GL.GL_COLOR_BUFFER_BIT, GL.GL_NEAREST);
		
		gl.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, 0);
		gl.glReadBuffer(GL.GL_BACK);
	}

	public boolean isDisposed() 
	{
		return framebufferId == 0;
	}
}
