package org.processmining.visualisation3d.gldatastructures;

import java.nio.FloatBuffer;

import org.processmining.visualisation3d.GraphicsPipeline;
import org.processmining.visualisation3d.maths.JoglMatrixd4x4;
import org.processmining.visualisation3d.maths.JoglVectord2;
import org.processmining.visualisation3d.maths.JoglVectord3;
import org.processmining.visualisation3d.maths.JoglVectord4;
import org.processmining.visualisation3d.settings.JoglGraphicsSettings;

import com.jogamp.opengl.GL2;
import com.jogamp.opengl.GL2ES2;
import com.jogamp.opengl.GLUniformData;
import com.jogamp.opengl.util.glsl.ShaderCode;
import com.jogamp.opengl.util.glsl.ShaderProgram;
import com.jogamp.opengl.util.glsl.ShaderState;

public class JoglShader implements IGLResource
{
	public static JoglShader boundShader = null;
	
	private boolean bLoaded = false;
	private ShaderState st;
	private String rootDir = "/org/processmining/processprofiler3d/compviz/jogl/";
	private String vsSourceFile = "textShaderVS";
	private String fsSourceFile = "textShaderFS";
	
	public JoglShader(String rDir, String vsFile, String fsFile)
	{
		rootDir = rDir;
		vsSourceFile = vsFile;
		fsSourceFile = fsFile;
	}
	
	public void Create(GraphicsPipeline pipe)
	{
		load(pipe);
	}

	@Override
	public boolean isLoaded()
	{
		return bLoaded;
	}
	
	@Override
	public void load(GraphicsPipeline pipe)
	{
		GL2 gl = pipe.gl;
		
		if (JoglGraphicsSettings.SupportsShaderObjects(pipe))
		{
	        st = new ShaderState();
	        
	        //st.setVerbose(true);
	        final ShaderCode vp0 = ShaderCode.create(gl, GL2ES2.GL_VERTEX_SHADER, this.getClass(), rootDir,
	                rootDir + "/bin", vsSourceFile, true);
	        final ShaderCode fp0 = ShaderCode.create(gl, GL2ES2.GL_FRAGMENT_SHADER, this.getClass(), rootDir,
	                rootDir + "/bin", fsSourceFile, true);        
	        
	        vp0.defaultShaderCustomization(gl, true, true);
	        fp0.defaultShaderCustomization(gl, true, true);
	        final ShaderProgram sp0 = new ShaderProgram();
	        sp0.add(gl, vp0, System.err);
	        sp0.add(gl, fp0, System.err);
	        st.attachShaderProgram(gl, sp0, true);
	        
	        if (!st.linked())
	        {
	        	System.out.println("Shader is not linked!");
	        }

	        //unbind shader to prevent it from polluting current pipelinestate
        	st.useProgram(gl, false);

        	if (pipe.gc != null)
        	{
        		pipe.gc.Track(this);
        	}
	        bLoaded = true;
		}
		
	}

	@Override
	public void unload(GraphicsPipeline pipe)
	{
		if (st != null)
		{
			st.destroy(pipe.gl);
			st = null;
		}

		if (pipe.gc != null)
		{
			pipe.gc.Untrack(this);
		}
		bLoaded = false;
	}
	
	public int getProgramHandle()
	{
		if (st != null)
		{
			return st.shaderProgram().id();
		}
		return 0;
	}
	
	public ShaderState getShaderState()
	{
		return st;
	}
	
	public void Bind(GraphicsPipeline pipe)
	{
		if (st != null)
		{
			st.useProgram(pipe.gl, true);
			boundShader = this;
		}
	}
	
	public void Unbind(GraphicsPipeline pipe)
	{
		if (st != null)
		{
			st.useProgram(pipe.gl, false);
			boundShader = null;
		}
	}
	
	public void SetUniform(GraphicsPipeline pipe, String uniformName, float f)
	{
		if (st != null)
		{
			GLUniformData myUniform = st.getUniform(uniformName);
			if (myUniform == null)
			{
				myUniform = new GLUniformData(uniformName, f);
				st.ownUniform(myUniform);
				st.uniform(pipe.gl, myUniform);
			}
			else
			{
				myUniform.setData(f);
				st.uniform(pipe.gl, myUniform);
			}
		}
	}
	
	public void SetUniform(GraphicsPipeline pipe, String uniformName, JoglVectord2 v)
	{
		if (st != null)
		{
			GLUniformData myUniform = st.getUniform(uniformName);
			if (myUniform == null)
			{
				FloatBuffer fb = FloatBuffer.allocate(2);
				fb.put((float)v.getValue(0));
				fb.put((float)v.getValue(1));
				fb.position(0);
				
				myUniform = new GLUniformData(uniformName, 2, fb);
				st.ownUniform(myUniform);
				//st.uniform(pipe.gl, myUniform);
			}
			else
			{
				FloatBuffer fb = myUniform.floatBufferValue();
				fb.position(0);
				fb.put((float)v.getValue(0));
				fb.put((float)v.getValue(1));
				fb.position(0);
				
				myUniform.setData(fb);
				//st.uniform(pipe.gl, myUniform);
			}
			st.uniform(pipe.gl, myUniform);
		}
	}
	
	public void SetUniform(GraphicsPipeline pipe, String uniformName, JoglVectord3 v)
	{
		if (st != null)
		{
			GLUniformData myUniform = st.getUniform(uniformName);
			if (myUniform == null)
			{
				FloatBuffer fb = FloatBuffer.allocate(3);
				fb.put((float)v.getValue(0));
				fb.put((float)v.getValue(1));
				fb.put((float)v.getValue(2));
				fb.position(0);
				
				myUniform = new GLUniformData(uniformName, 3, fb);
				st.ownUniform(myUniform);
	        	//st.uniform(pipe.gl, myUniform);
			}
			else
			{
				FloatBuffer fb = myUniform.floatBufferValue();
				fb.position(0);
				fb.put((float)v.getValue(0));
				fb.put((float)v.getValue(1));
				fb.put((float)v.getValue(2));
				fb.position(0);
				
				myUniform.setData(fb);
	        	//st.uniform(pipe.gl, myUniform);
			}
			st.uniform(pipe.gl, myUniform);
		}
	}
	
	public void SetUniform(GraphicsPipeline pipe, String uniformName, JoglVectord4 v)
	{
		if (st != null)
		{
			GLUniformData myUniform = st.getUniform(uniformName);
			if (myUniform == null)
			{
				FloatBuffer fb = FloatBuffer.allocate(4);
				fb.put((float)v.getValue(0));
				fb.put((float)v.getValue(1));
				fb.put((float)v.getValue(2));
				fb.put((float)v.getValue(3));
				fb.position(0);
				
				myUniform = new GLUniformData(uniformName, 4, fb);
				st.ownUniform(myUniform);
				//st.uniform(pipe.gl, myUniform);
			}
			else
			{
				FloatBuffer fb = myUniform.floatBufferValue();
				fb.position(0);
				fb.put((float)v.getValue(0));
				fb.put((float)v.getValue(1));
				fb.put((float)v.getValue(2));
				fb.put((float)v.getValue(3));
				fb.position(0);
				
				myUniform.setData(fb);
			}
			st.uniform(pipe.gl, myUniform);
		}
	}

	public void SetUniformMatrix(GraphicsPipeline pipe, String uniformName, JoglMatrixd4x4 mat)
	{
//		if (st != null)
//		{
//			GLUniformData myUniform = st.getUniform(uniformName);
//			if (myUniform == null)
//			{
//				FloatBuffer fb = FloatBuffer.allocate(16);
//				double[] mdata = mat.GetData();
//				fb.put((float)mdata[0]);
//				fb.put((float)mdata[1]);
//				fb.put((float)mdata[2]);
//				fb.put((float)mdata[3]);
//				fb.put((float)mdata[4]);
//				fb.put((float)mdata[5]);
//				fb.put((float)mdata[6]);
//				fb.put((float)mdata[7]);
//				fb.put((float)mdata[8]);
//				fb.put((float)mdata[9]);
//				fb.put((float)mdata[10]);
//				fb.put((float)mdata[11]);
//				fb.put((float)mdata[12]);
//				fb.put((float)mdata[13]);
//				fb.put((float)mdata[14]);
//				fb.put((float)mdata[15]);
//				fb.position(0);
//				
//				myUniform = new GLUniformData(uniformName, 4, fb);
//				st.ownUniform(myUniform);
//				//st.uniform(pipe.gl, myUniform);
//			}
//			else
//			{
//				FloatBuffer fb = myUniform.floatBufferValue();
//				fb.position(0);
//				double[] mdata = mat.GetData();
//				fb.put((float)mdata[0]);
//				fb.put((float)mdata[1]);
//				fb.put((float)mdata[2]);
//				fb.put((float)mdata[3]);
//				fb.put((float)mdata[4]);
//				fb.put((float)mdata[5]);
//				fb.put((float)mdata[6]);
//				fb.put((float)mdata[7]);
//				fb.put((float)mdata[8]);
//				fb.put((float)mdata[9]);
//				fb.put((float)mdata[10]);
//				fb.put((float)mdata[11]);
//				fb.put((float)mdata[12]);
//				fb.put((float)mdata[13]);
//				fb.put((float)mdata[14]);
//				fb.put((float)mdata[15]);
//				fb.position(0);
//				
//				myUniform.setData(fb);
//			}
//			st.uniform(pipe.gl, myUniform);
//		}		
		
		int loc = st.getUniformLocation(pipe.gl, uniformName);
		double[] mdata = mat.GetData();
		float[] fdata = new float[mdata.length];
		for (int x = 0; x < mdata.length; x++)
		{
			fdata[x] = (float)mdata[x];
		}
		pipe.gl.glUniformMatrix4fv(loc, 1, false, fdata, 0);
	}
	
	public void Dispose(GraphicsPipeline pipe)
	{
		unload(pipe);
	}

	public boolean isDisposed() 
	{
		return st == null;
	}
}
