package org.processmining.visualisation3d.gldatastructures;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.ByteBuffer;

import org.processmining.visualisation3d.GraphicsPipeline;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2;
import com.jogamp.opengl.GL2ES3;
import com.jogamp.opengl.GL2GL3;
import com.jogamp.opengl.GLException;
import com.jogamp.opengl.util.GLPixelStorageModes;
import com.jogamp.opengl.util.texture.Texture;
import com.jogamp.opengl.util.texture.TextureData;
import com.jogamp.opengl.util.texture.TextureIO;
import com.jogamp.opengl.util.texture.awt.AWTTextureIO;

public class JoglTexture implements IGLResource
{
	protected boolean bLoaded = false;
	protected Texture tex;
	
	protected TextureData tdata = null;
	protected int minFilter = 0;
	protected int magFilter = 0;
	protected int anisoSamples = 0;
	
	public int getTextureId()
	{
		if (tex != null)
		{
			return tex.getTextureObject();
		}
		return 0;
	}
	
	public Texture getTex()
	{
		return tex;
	}

	protected JoglTexture()
	{
		
	}
	
	public JoglTexture(GraphicsPipeline pipe, Texture t)
	{
		tex = t;
		
		bLoaded = true;
		if (pipe.gc != null)
		{
			pipe.gc.Track(this);
		}
	}
	
	public static JoglTexture LoadTextureFromFile(GraphicsPipeline pipe, String texResLoc, boolean bMipMap)
	{
		GL gl = pipe.gl;
		
		System.out.println("Loading texture: " + texResLoc);
		URL fileUrl = pipe.getClass().getResource(texResLoc);

		String texFileType = DetermineFileTypeHint(texResLoc);
		
		try
		{
			TextureData td = TextureIO.newTextureData(gl.getGLProfile(), fileUrl, bMipMap, texFileType);
			Texture tx = TextureIO.newTexture(td);
			
			JoglTexture jt = new JoglTexture(pipe, tx);
			jt.setDefaultFilterSettings(pipe, bMipMap);
			
			return jt;
		} 
		catch (IOException e) 
		{
			// TODO Auto-generated catch block
			//e.printStackTrace();
			System.out.println(e.getMessage());
		}
		return null;
	}
	
	public static JoglTexture LoadTextureFromFile(GraphicsPipeline pipe, URL fileUrl, boolean bMipMap)
	{
		GL gl = pipe.gl;
		
		System.out.println("Loading texture: " + fileUrl.getPath());

		String texFileType = DetermineFileTypeHint(fileUrl.getPath());
		
		try
		{
			TextureData td = TextureIO.newTextureData(gl.getGLProfile(), fileUrl, bMipMap, texFileType);
			Texture tx = TextureIO.newTexture(td);
			
			JoglTexture jt = new JoglTexture(pipe, tx);
			jt.setDefaultFilterSettings(pipe, bMipMap);
			
			return jt;
		} 
		catch (IOException e) 
		{
			// TODO Auto-generated catch block
			//e.printStackTrace();
			System.out.println(e.getMessage());
		}
		return null;
	}
	
	public static JoglTexture LoadTextureFromImage(GraphicsPipeline pipe, BufferedImage img, boolean bMipMap)
	{

		GL gl = pipe.gl;
		
		System.out.println("Loading texture from image.");
		
		TextureData td = AWTTextureIO.newTextureData(gl.getGLProfile(), img, bMipMap);
		Texture tx = TextureIO.newTexture(td);
		
		JoglTexture jt = new JoglTexture(pipe, tx);
		jt.setDefaultFilterSettings(pipe, bMipMap);
		
		return jt;
	}

	protected static String DetermineFileTypeHint(String texResLoc)
	{
		String texResLocLC = texResLoc.toLowerCase();
		String texFileType = "";
		if (texResLocLC.endsWith(".tga"))
		{
			texFileType = TextureIO.TGA;
		}
		else if (texResLocLC.endsWith(".png"))
		{
			texFileType = TextureIO.PNG;
		}
		else if (texResLocLC.endsWith(".jpg"))
		{
			texFileType = TextureIO.JPG;
		}
		else if (texResLocLC.endsWith(".gif"))
		{
			texFileType = TextureIO.GIF;
		}
		else if (texResLocLC.endsWith(".dds"))
		{
			texFileType = TextureIO.DDS;
		}
		else
		{
			System.out.println("Trying to load texture from unknown file type!");
			texFileType = null;
		}
		return texFileType;
	}
	
	public void setDefaultFilterSettings(GraphicsPipeline pipe, boolean isMipMapped)
	{
		GL gl = pipe.gl;
		
		if (isMipMapped)
		{
			minFilter = GL.GL_LINEAR_MIPMAP_LINEAR;
		}
		else
		{
			minFilter = GL.GL_LINEAR;
		}
		magFilter = GL.GL_LINEAR;
		anisoSamples = pipe.getGraphicsSettings().GetAnisotropicSamples();		
		
		tex.enable(gl);
		tex.bind(gl);
		tex.setTexParameteri(gl, GL.GL_TEXTURE_MIN_FILTER, minFilter);
		tex.setTexParameteri(gl, GL.GL_TEXTURE_MAG_FILTER, magFilter);
		tex.setTexParameteri(gl, GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, anisoSamples);
		gl.glBindTexture(GL.GL_TEXTURE_2D, 0);
		
	}
	
	@Override
	public boolean isLoaded()
	{
		return bLoaded;
	}
	
	@Override
	public void load(GraphicsPipeline pipe)
	{
		GL gl = pipe.gl;
		
		if (tdata != null)
		{
			tex = TextureIO.newTexture(tdata);
			tex.enable(gl);
			tex.bind(gl);
			tex.setTexParameteri(gl, GL.GL_TEXTURE_MIN_FILTER, minFilter);
			tex.setTexParameteri(gl, GL.GL_TEXTURE_MAG_FILTER, magFilter);
			tex.setTexParameteri(gl, GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, anisoSamples);

			if (pipe.gc != null)
			{
				pipe.gc.Track(this);
			}
			bLoaded = true;
		}
	}

	@Override
	public void unload(GraphicsPipeline pipe)
	{
		try
		{
			GL2 gl = pipe.gl;
			
			tdata = getTextureData(pipe);
			tex.destroy(gl);
			tex = null;
			
			//should maybe get current tex filter settings (in case someone updated them without using the proper functions)?
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
		}
		if (pipe.gc != null)
		{
			pipe.gc.Untrack(this);
		}
		bLoaded = false;
	}
	
	public void writeToFile(String texResLoc)
	{
		try 
		{
			TextureIO.write(tex, new File(texResLoc));
		} 
		catch (GLException e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		} 
		catch (IOException e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}	
	}
	
	public TextureData getTextureData(GraphicsPipeline pipe) throws IOException, GLException 
	{
		GL2 gl = pipe.gl;
		
		final Texture texture = tex;
        if (texture.getTarget() != GL.GL_TEXTURE_2D) 
        {
            throw new GLException("Only GL_TEXTURE_2D textures are supported");
        }

        texture.bind(gl);
        final int internalFormat = glGetTexLevelParameteri(gl, GL.GL_TEXTURE_2D, 0, GL2ES3.GL_TEXTURE_INTERNAL_FORMAT);
        final int width  = glGetTexLevelParameteri(gl, GL.GL_TEXTURE_2D, 0, GL2ES3.GL_TEXTURE_WIDTH);
        final int height = glGetTexLevelParameteri(gl, GL.GL_TEXTURE_2D, 0, GL2ES3.GL_TEXTURE_HEIGHT);
        final int border = glGetTexLevelParameteri(gl, GL.GL_TEXTURE_2D, 0, GL2.GL_TEXTURE_BORDER);
        TextureData data = null;
        if (internalFormat == GL.GL_COMPRESSED_RGB_S3TC_DXT1_EXT ||
            internalFormat == GL.GL_COMPRESSED_RGBA_S3TC_DXT1_EXT ||
            internalFormat == GL.GL_COMPRESSED_RGBA_S3TC_DXT3_EXT ||
            internalFormat == GL.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT) {
            // Fetch using glGetCompressedTexImage
            final int size   = glGetTexLevelParameteri(gl, GL.GL_TEXTURE_2D, 0, GL2GL3.GL_TEXTURE_COMPRESSED_IMAGE_SIZE);
            final ByteBuffer res = ByteBuffer.allocate(size);
            gl.glGetCompressedTexImage(GL.GL_TEXTURE_2D, 0, res);
            data = new TextureData(gl.getGLProfile(), internalFormat, width, height, border, internalFormat, GL.GL_UNSIGNED_BYTE,
                                   false, true, true, res, null);
        } 
        else 
        {
            int bytesPerPixel = 0;
            int fetchedFormat = 0;
            switch (internalFormat) 
            {
            case GL.GL_RGB:
            case GL.GL_BGR:
            case GL.GL_RGB8:
                bytesPerPixel = 3;
                fetchedFormat = GL.GL_RGB;
                break;
            case GL.GL_RGBA:
            case GL.GL_BGRA:
            case GL2.GL_ABGR_EXT:
            case GL.GL_RGBA8:
                bytesPerPixel = 4;
                fetchedFormat = GL.GL_RGBA;
                break;
            default:
                throw new IOException("Unsupported texture internal format 0x" + Integer.toHexString(internalFormat));
            }

            // Fetch using glGetTexImage
            final GLPixelStorageModes psm = new GLPixelStorageModes();
            psm.setPackAlignment(gl, 1);

            final ByteBuffer res = ByteBuffer.allocate((width + (2 * border)) *
                                                 (height + (2 * border)) *
                                                 bytesPerPixel);
            
            gl.glGetTexImage(GL.GL_TEXTURE_2D, 0, fetchedFormat, GL.GL_UNSIGNED_BYTE, res);

            psm.restore(gl);

            data = new TextureData(gl.getGLProfile(), internalFormat, width, height, border, fetchedFormat, GL.GL_UNSIGNED_BYTE,
                                   true, false, false, res, null);

        }

        return data;
    }
	
	private static int glGetTexLevelParameteri(final GL2 gl, final int target, final int level, final int pname) 
	{
        final int[] tmp = new int[1];
        gl.glGetTexLevelParameteriv(target, 0, pname, tmp, 0);
        return tmp[0];
    }
    
	public void Bind(GraphicsPipeline pipe)
	{
		tex.bind(pipe.gl);
	}
	
	public void Unbind(GraphicsPipeline pipe)
	{
		pipe.gl.glBindTexture(GL.GL_TEXTURE_2D, 0);
	}
	
	public void setAnisoSamples(GraphicsPipeline pipe, int numSamples)
	{
		tex.setTexParameteri(pipe.gl, GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, numSamples);
	}
	
	public void Dispose(GraphicsPipeline pipe) 
	{
		if (tex != null)
		{
			tex.destroy(pipe.gl);
			tex = null;

			if (pipe.gc != null)
			{
				pipe.gc.Untrack(this);
			}
			bLoaded = false;
		}
	}

	public boolean isDisposed() 
	{
		return tex == null;
	}

}
