package org.processmining.visualisation3d.gldatastructures;

import java.nio.IntBuffer;
import java.util.Vector;

import org.processmining.visualisation3d.GraphicsPipeline;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2;

public class JoglVertexArrayObject implements IGLResource
{
	private boolean bLoaded = false;
	private int vaoHandle = 0;
	
	private Vector<IJoglVertexArrayObjectBuffer> buffers = new Vector<IJoglVertexArrayObjectBuffer>();
	private IJoglVertexArrayObjectIndexBuffer indbuffer;
	
	private static int GenVertexArrayObject(GL2 gl)
	{
		IntBuffer buffer = IntBuffer.allocate(1);
		gl.glGenVertexArrays(1, buffer);
		return buffer.get(0);	
	}
	
	private static void DeleteVertexArrayObject(GL2 gl, int id)
	{
		IntBuffer buffer = IntBuffer.allocate(1);
		buffer.put(id);
		buffer.rewind();
		gl.glDeleteVertexArrays(1, buffer);	
	}
	
	public JoglVertexArrayObject(GraphicsPipeline pipe)
	{
        load(pipe);
	}
    
	@Override
	public boolean isLoaded()
	{
		return bLoaded;
	}
	
    @Override
    public void load(GraphicsPipeline pipe)
    {
    	vaoHandle = GenVertexArrayObject(pipe.gl);

        for (IJoglVertexArrayObjectBuffer buffer : buffers)
        {
    		if (!buffer.isLoaded())
    		{
    			buffer.load(pipe);
    		}
        }
        
        if (indbuffer != null)
        {
    		if (!indbuffer.isLoaded())
    		{
    			indbuffer.load(pipe);
    		}
        }
        
        Bind(pipe);
    	UpdateBufferAttachment(pipe);
    	Unbind(pipe);

    	if (pipe.gc != null)
    	{
    		pipe.gc.Track(this);
    	}
    	bLoaded = true;
    }
    
    @Override
    public void unload(GraphicsPipeline pipe)
    {
    	//deattached buffers
    	Bind(pipe);
        int i = 0;
        pipe.gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
        for (IJoglVertexArrayObjectBuffer buffer : buffers)
        {
            pipe.gl.glVertexAttribPointer(i, 1, GL.GL_FLOAT, true, 4, 0);
            pipe.gl.glDisableVertexAttribArray(i);
            i++;
        }
        Unbind(pipe);
    	
    	DeleteVertexArrayObject(pipe.gl, vaoHandle);
		vaoHandle = 0;

		if (pipe.gc != null)
		{
			pipe.gc.Untrack(this);
		}
    	bLoaded = false;
    }

	public void Dispose(GL2 gl)
	{
        GraphicsPipeline pipe = new GraphicsPipeline();
        pipe.gl = gl;
        
        unload(pipe);

        for (IJoglVertexArrayObjectBuffer buffer : buffers)
        {
        	if (buffer.isLoaded())
        	{
        		buffer.unload(pipe);
        	}
        }
        
        if (indbuffer.isLoaded())
        {
        	indbuffer.unload(pipe);
        }
	}
	
	public Vector<IJoglVertexArrayObjectBuffer> getBuffers()
	{
		return buffers;
	}
	
    public void AttachBuffer(IJoglVertexArrayObjectBuffer buffer)
    {
        buffers.add(buffer);
    }
	
    public void AttachIndexBuffer(IJoglVertexArrayObjectIndexBuffer ibuffer)
    {
        indbuffer = ibuffer;
    }
    
    public void ClearBuffers()
    {
    	buffers.clear();
    }

    public void UpdateBufferAttachment(GraphicsPipeline pipe)
    {
    	if (buffers.isEmpty())
    	{
    		return;
    	}
    	
        //Bind();
        int i = 0;
        for (IJoglVertexArrayObjectBuffer buffer : buffers)
        {
            buffer.Attach(pipe, i);
            i++;
        }
        indbuffer.Attach(pipe);
        buffers.get(0).Unbind(pipe);
        //Unbind();
    }

    public void Bind(GraphicsPipeline pipe)
    {
        pipe.gl.glBindVertexArray(vaoHandle);
    }

    public void Unbind(GraphicsPipeline pipe)
    {
        pipe.gl.glBindVertexArray(0);
    }

	public boolean isDisposed() 
	{
		return vaoHandle == 0;
	}
}
