package org.processmining.visualisation3d.graphicsdatastructures;

import java.util.Vector;

import org.processmining.visualisation3d.GraphicsPipeline;
import org.processmining.visualisation3d.collision.JoglAABB;
import org.processmining.visualisation3d.collision.JoglRay;
import org.processmining.visualisation3d.maths.JoglMatrixd4x4;
import org.processmining.visualisation3d.maths.JoglVectord2;
import org.processmining.visualisation3d.maths.JoglVectord3;

public class JoglCamera 
{
	private JoglVectord3 position;
	private JoglVectord3 lookAt;
	private JoglVectord3 upDir;
	
	private JoglMatrixd4x4 rotationMatrix;
	
	public JoglCamera()
	{	
		position = new JoglVectord3(0,0,-5.0);
		lookAt = new JoglVectord3(0,0,0);
		upDir = new JoglVectord3(0,1,0);

		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public JoglVectord3 GetPosition()
	{
		return position;
	}
	
	public void SetPosition(JoglVectord3 v)
	{
		position = v;
		
		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public JoglVectord3 GetLookAt()
	{
		return lookAt;
	}
	
	public JoglVectord3 GetLookDir()
	{
		return new JoglVectord3(-rotationMatrix.GetValue(2, 0), -rotationMatrix.GetValue(2, 1), -rotationMatrix.GetValue(2, 2));
	}
	
	public JoglVectord3 GetLeftDir()
	{
    	return new JoglVectord3(-rotationMatrix.GetValue(0, 0), -rotationMatrix.GetValue(0, 1), -rotationMatrix.GetValue(0, 2));
	}
	
	public void SetLookAt(JoglVectord3 v)
	{
		lookAt = v;
		
		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public void SetLookAt(JoglVectord3 v, JoglVectord3 u)
	{
		lookAt = v;
		upDir = u;
		
		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public void SetLookDir(JoglVectord3 v)
	{
		lookAt = JoglVectord3.Add(this.GetPosition(), v);
		
		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public void SetLookDir(JoglVectord3 v, JoglVectord3 u)
	{
		lookAt = JoglVectord3.Add(this.GetPosition(), v);
		upDir = u;
		
		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public JoglVectord3 GetUpDir()
	{
		return new JoglVectord3(rotationMatrix.GetValue(1, 0), rotationMatrix.GetValue(1, 1), rotationMatrix.GetValue(1, 2));
	}
	
	public void SetUpDir(JoglVectord3 v)
	{
		upDir = v;
		
		rotationMatrix = JoglMatrixd4x4.CreateRotationFromLookAt(position, lookAt, upDir);
	}
	
	public double GetPitch()
	{
		//return Math.asin(rotationMatrix.GetValue(2, 3));
    	double dotx = GetLookDir().getY();
    	double pitch = -Math.asin(dotx);

    	return pitch;
	}
	
	public double GetYaw()
	{
		//return Math.atan2(rotationMatrix.GetValue(1, 3), rotationMatrix.GetValue(3, 3));
    	double doty = GetLeftDir().getZ();
    	double yaw = -Math.asin(doty);

    	if (GetLeftDir().getX() < 0)
    	{
    		return yaw;
    	}
    	else
    	{
    		double cyaw = Math.PI - yaw;
    		return cyaw;
    	}
	}
	
	public void SetPitchYawRoll(double p, double y, double r)
	{
		rotationMatrix = JoglMatrixd4x4.CreateRotationPYR(p,y,r);
		
		lookAt = JoglVectord3.Add(GetPosition(), GetLookDir());
		upDir = GetUpDir();
	}
	
	public JoglMatrixd4x4 GetCameraMatrix()
	{
		JoglMatrixd4x4 cmat = JoglMatrixd4x4.CreateTranslation(-position.getValue(0), -position.getValue(1), -position.getValue(2));
		JoglMatrixd4x4 tmat = JoglMatrixd4x4.Multiply(rotationMatrix, cmat);
		
		if (lens instanceof JoglCameraLensOrthographic)
		{
			double cdist = 1000.0f / position.getMagnitude();
			double camZoom = cdist * cdist;
			camZoom = 1.0;
			
			JoglCameraLensOrthographic jclo = (JoglCameraLensOrthographic)lens;
			jclo.setZoom((float)camZoom);
			tmat = JoglMatrixd4x4.Multiply(JoglMatrixd4x4.CreateTranslation(0, 0, jclo.getScaleD() * camZoom), tmat);	
			tmat = JoglMatrixd4x4.Multiply(JoglMatrixd4x4.CreateScale(camZoom), tmat);
		}
		
		return tmat;
	}
	
	private JoglCameraLens lens = new JoglCameraLensPerspective(1.0f);
	
	public JoglCameraLens getCameraLens()
	{
		return lens;
	}
	
	public void setCameraLens(JoglCameraLens jcl)
	{
		lens = jcl;
	}
	
	public void DrawCamera(GraphicsPipeline pipe)
	{
		//double aspect = targetSize.getWidth() / targetSize.getHeight();
	    SetProjectionMatrix(pipe);

		SetCameraMatrix(pipe);

//		System.out.println("Camera:");
//		logCameraMatrix();
//
//		System.out.println("Projection:");
//		logProjectionMatrix();
	}

	public void SetCameraMatrix(GraphicsPipeline pipe) 
	{
		JoglMatrixd4x4 cmat = GetCameraMatrix();
//		double[] cdat = cmat.GetData();
//		gl.glLoadMatrixd(cdat, 0);
		pipe.transform.LoadMatrix(cmat);
	}

	public void SetProjectionMatrix(GraphicsPipeline pipe) 
	{
		lens.SetProjectionMatrix(pipe);
	}

	public JoglVectord2 calcDepthRange(JoglAABB aabb) 
	{
		Vector<JoglVectord3> lst = aabb.getCorners();
		
		double minDist = Double.POSITIVE_INFINITY;
		double maxDist = Double.NEGATIVE_INFINITY;
		
		for (JoglVectord3 jv : lst)
		{
			JoglVectord3 vdir = JoglVectord3.Subtract(jv, GetPosition());	
			double z = JoglVectord3.Dot(vdir, GetLookDir());
			
			minDist = Math.min(minDist, z);
			maxDist = Math.max(maxDist, z);
		}
		
		float ndist = (float)Math.max(0.01, minDist - 50.0);
		float fdist = (float)Math.max(ndist + 0.01, maxDist + 50.0);

		//System.out.println("Depth range: " + ndist + " - " + fdist);
		//setDepthRange(ndist, fdist);
		return new JoglVectord2(ndist, fdist);
	}

	public JoglVectord2 calcDepthRange(Vector<JoglAABB> lst) 
	{
		if (lst.isEmpty())
		{
			return null;
		}
		
		JoglVectord2 oldRange = calcDepthRange(lst.get(0));
		for (int i = 1; i < lst.size(); i++)
		{
			JoglVectord2 newRange = calcDepthRange(lst.get(i));
			
			oldRange = new JoglVectord2(Math.min(oldRange.getX(), newRange.getX()), Math.max(oldRange.getY(), newRange.getY()));
		}
		
		return oldRange;
	}

	public JoglVectord3 ProjectToScreen(JoglVectord3 v, double aspect)
	{
		JoglVectord3 vdir = JoglVectord3.Subtract(v, GetPosition());
		
		double x = JoglVectord3.Dot(vdir, GetLeftDir());
		double y = JoglVectord3.Dot(vdir, GetUpDir());
		double z = JoglVectord3.Dot(vdir, GetLookDir());
		
		return lens.ProjectToScreen(x, y, z, aspect);
	}
	
//	public void Unproject(GL2 gl, int mouseX, int mouseY)
//	{
//		GLU glu = new GLU();
//
//	    int viewport[] = new int[4];
//	    double mvmatrix[] = new double[16];
//	    double projmatrix[] = new double[16];
//	    int realy = 0;// GL y coord pos
//	    double wcoord[] = new double[4];// wx, wy, wz;// returned xyz coords
//	    
//	      int x = mouseX, y = mouseY;
//	      gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
//	      gl.glGetDoublev(GLMatrixFunc.GL_MODELVIEW_MATRIX, mvmatrix, 0);
//	      gl.glGetDoublev(GLMatrixFunc.GL_PROJECTION_MATRIX, projmatrix, 0);
//	      /* note viewport[3] is height of window in pixels */
//	      realy = viewport[3] - (int)y - 1;
//	      System.out.println("Coordinates at cursor are (" + x + ", " + realy);
//	      glu.gluUnProject((double) x, (double) realy, 0.0, //
//	          mvmatrix, 0,
//	          projmatrix, 0, 
//	          viewport, 0, 
//	          wcoord, 0);
//	      System.out.println("World coords at z=0.0 are ( " //
//	                         + wcoord[0] + ", " + wcoord[1] + ", " + wcoord[2]
//	                         + ")");
//	      glu.gluUnProject((double) x, (double) realy, 1.0, //
//	          mvmatrix, 0,
//	          projmatrix, 0,
//	          viewport, 0, 
//	          wcoord, 0);
//	      System.out.println("World coords at z=1.0 are (" //
//	                         + wcoord[0] + ", " + wcoord[1] + ", " + wcoord[2]
//	                         + ")");
//
//	}
    
    public JoglRay GetRay(JoglVectord2 p)
    {
		JoglRay r = lens.GetRay(this, p);
		return r;
    }
    
    public void logCameraMatrix()
    {
    	JoglMatrixd4x4 mat = GetCameraMatrix();
    	
    	System.out.println(JoglMatrixd4x4.print(mat));
    }
    
    public void logProjectionMatrix()
    {
    	if (lens != null)
    	{
    		lens.logProjectionMatrix();
    	}
    }

}
