package org.processmining.visualisation3d.graphicsdatastructures;

import org.processmining.visualisation3d.collision.JoglRay;
import org.processmining.visualisation3d.maths.JoglMatrixd4x4;
import org.processmining.visualisation3d.maths.JoglVectord2;
import org.processmining.visualisation3d.maths.JoglVectord3;

public class JoglCameraLensPerspective extends JoglCameraLens 
{
	private float fieldOfView = 45.0f;
	private double tanFoV = Math.tan(45.0 * 0.5 * 0.0174532925);
	public double getFieldOfView()
	{
		return fieldOfView;
	}
	
	public double getTanFoV()
	{
		return tanFoV;
	}

	public JoglCameraLensPerspective(double aspect)
	{
		this.aspect = aspect;
		projMatrix = GetProjectionMatrixOpt(aspect);
	}

	public JoglCameraLensPerspective(double fov, double aspect)
	{
		this.aspect = aspect;
		fieldOfView = (float)fov;
		tanFoV = Math.tan(fieldOfView * 0.5 * 0.0174532925);
		
		projMatrix = GetProjectionMatrixOpt(aspect);
	}

	public JoglCameraLensPerspective(double fov, double aspect, double nDist, double fDist)
	{
		this.aspect = aspect;
		fieldOfView = (float)fov;
		tanFoV = Math.tan(fieldOfView * 0.5 * 0.0174532925);
		nearDist = (float)nDist;
		farDist = (float)fDist;
		
		projMatrix = GetProjectionMatrixOpt(aspect);		
	}

	public JoglCameraLensPerspective(double fov, double aspect, double nDist, double fDist, double l, double r, double b, double t)
	{
		this.aspect = aspect;
		fieldOfView = (float)fov;
		tanFoV = Math.tan(fieldOfView * 0.5 * 0.0174532925);
		nearDist = (float)nDist;
		farDist = (float)fDist;
		
		projMatrix = GetProjectionMatrixOpt(aspect, l, r, b, t);	
	}

	public JoglCameraLensPerspective(double fov, double aspect, JoglMatrixd4x4 mat, double nDist, double fDist)
	{
		this.aspect = aspect;
		fieldOfView = (float)fov;
		tanFoV = Math.tan(fieldOfView * 0.5 * 0.0174532925);
		nearDist = (float)nDist;
		farDist = (float)fDist;
		
		projMatrix = mat;	
	}
	
	@Override
	public JoglVectord3 ProjectToScreen(double x, double y, double z, double aspect)
	{
		double factor = (tanFoV * z);
		double perspectiveFactorX = 1.0 / (factor * aspect);
		double perspectiveFactorY = 1.0 / (factor);
		
		return new JoglVectord3(x * perspectiveFactorX, y * perspectiveFactorY, z);
	}
	
	private JoglMatrixd4x4 GetProjectionMatrixOpt(double aspect) 
	{
		//adapted from http://nehe.gamedev.net/article/replacement_for_gluperspective/21002/
		double fH = tanFoV * nearDist;
		double fW = fH * aspect;

		JoglMatrixd4x4 mat = JoglMatrixd4x4.CreateFrustum(nearDist, farDist, -fW, fW, -fH, fH);
		return mat;
	}

	private JoglMatrixd4x4 GetProjectionMatrixOpt(double aspect, double left, double right, double top, double bottom) 
	{
		//adapted from http://nehe.gamedev.net/article/replacement_for_gluperspective/21002/
		double fH = tanFoV * nearDist;
		double fW = fH * aspect;

		double w = fW * 2.0;
		double h = fH * 2.0;
		
		double l = -fW;
		//double r = fW;
		double b = -fH;
		//double t = fH;
		
		JoglMatrixd4x4 mat = JoglMatrixd4x4.CreateFrustum(nearDist, farDist, l + (w * left), l + (w * right), b + (h * bottom), b + (h * top));
		return mat;
	}

    public JoglVectord3 GetRayHelper(float aspect_ratio, float px, float py)
    {
        float near = nearDist;

        float opp2 = (float)tanFoV * near;
        float opp1 = opp2 * aspect_ratio;

        float totalWidth = opp1 * 2.0f;
        float totalHeight = -(opp2 * 2.0f);

        float totalLeft = -opp1;
        float totalTop = opp2;

        float left = totalLeft;
        float right = totalLeft + totalWidth;
        float top = totalTop;
        float bottom = totalTop + totalHeight;

        float w = right - left;
        float h = bottom - top;

        return new JoglVectord3(left + (w * px), top + (h * py), near);
    }

    public JoglVectord3 GetRayDir(JoglCamera cam, JoglVectord2 p)
    {
        //float aspect = (float)viewWidth / (float)viewHeight;

        float px = (float)p.getX();
        float py = (float)p.getY();

        //ray position in camera space
        JoglVectord3 pixelPos = GetRayHelper((float)aspect, px, py);

        //transform ray position into world space
        JoglVectord3 rayForward = JoglVectord3.Multiply(cam.GetLookDir(), pixelPos.getZ());
        JoglVectord3 rayLeft = JoglVectord3.Multiply(cam.GetLeftDir(), -pixelPos.getX());
        JoglVectord3 rayUp = JoglVectord3.Multiply(cam.GetUpDir(), pixelPos.getY());
        
        JoglVectord3 rayDir = JoglVectord3.Add(JoglVectord3.Add(rayForward, rayLeft), rayUp);
        rayDir = JoglVectord3.Normal(rayDir);

        return rayDir;
    }
    
    @Override
    public JoglRay GetRay(JoglCamera cam, JoglVectord2 p)
    {
    	JoglVectord3 camPos = cam.GetPosition();  	
		JoglRay r = new JoglRay(camPos, GetRayDir(cam, p));
		return r;
    }
}
