package org.processmining.visualisation3d.maths;

public class JoglMatrixd4x4
{
	private double[] data;
	
	public JoglMatrixd4x4()
	{
		data = new double[16];
	}
	
	public double[] GetData()
	{
		return data;
	}
	
	public void SetData(double[] d)
	{
		data = d;
	}
	
	public double GetValue(int r, int c)
	{
		return data[(c * 4) + r];
	}
	
	public void SetValue(int r, int c, double v)
	{
		data[(c*4) + r] = v;
	}

	public static JoglMatrixd4x4 Multiply(JoglMatrixd4x4 m1, JoglMatrixd4x4 m2)
	{
    	JoglMatrixd4x4 nm = new JoglMatrixd4x4();

        for (int j = 0; j < 4; j++)
        {
            for (int i = 0; i < 4; i++)
            {
                for (int k = 0; k < 4; k++)
                {
                    nm.SetValue(j, i, nm.GetValue(j, i) + m1.GetValue(j, k) * m2.GetValue(k, i));
                }
            }
        }
        
    	return nm;
	}
	
    public static JoglMatrixd4x4 CreateTranspose(JoglMatrixd4x4 om)
    {
    	JoglMatrixd4x4 nm = new JoglMatrixd4x4();

        for (int j = 0; j < 4; j++)
        {
            for (int i = 0; i < 4; i++)
            {
                nm.SetValue(i, j, om.GetValue(j,i));
            }
        }

        return nm;
    }
	
	public static JoglMatrixd4x4 CreateIdentity()
	{
		JoglMatrixd4x4 mat = new JoglMatrixd4x4();

        for (int i = 0; i < 4; i++)
        {
            mat.SetValue(i, i, 1);
        }
        
        return mat;
	}

    //roll
    public static JoglMatrixd4x4 CreateRotationX(double angle)
    {
    	JoglMatrixd4x4 nm = CreateIdentity();

        nm.SetValue(1, 1, Math.cos(angle));
        nm.SetValue(1, 2, -Math.sin(angle));
        nm.SetValue(2, 1, -nm.GetValue(1, 2));
        nm.SetValue(2, 2, nm.GetValue(1, 1));

        return nm;
    }

    //pitch
    public static JoglMatrixd4x4 CreateRotationY(double angle)
    {
    	JoglMatrixd4x4 nm = CreateIdentity();

        nm.SetValue(0, 0, Math.cos(angle));
        nm.SetValue(0, 2, Math.sin(angle));
        nm.SetValue(2, 0, -nm.GetValue(0, 2));
        nm.SetValue(2, 2, nm.GetValue(0, 0));

        return nm;
    }

    //yaw
    public static JoglMatrixd4x4 CreateRotationZ(double angle)
    {
    	JoglMatrixd4x4 nm = CreateIdentity();

        nm.SetValue(0, 0, Math.cos(angle));
        nm.SetValue(0, 1, -Math.sin(angle));
        nm.SetValue(1, 0, -nm.GetValue(0, 1));
        nm.SetValue(1, 1, nm.GetValue(0, 0));

        return nm;
    }

    //create camera rotation matrix from pitch, yaw and roll
    public static JoglMatrixd4x4 CreateRotationPYR(double p, double y, double r)
    {
        //return JoglMatrixd4x4.Multiply(JoglMatrixd4x4.Multiply(JoglMatrixd4x4.CreateRotationY(y), JoglMatrixd4x4.CreateRotationX(p)), JoglMatrixd4x4.CreateRotationZ(r));
        return JoglMatrixd4x4.Multiply(JoglMatrixd4x4.CreateRotationX(p), JoglMatrixd4x4.CreateRotationY(y));
    }

    //create camera rotation matrix given a position the camera should be focused on
    public static JoglMatrixd4x4 CreateRotationFromLookAt(JoglVectord3 p, JoglVectord3 la, JoglVectord3 u)
    {
    	JoglVectord3 lookDir = JoglVectord3.Subtract(la, p);
    	JoglVectord3 udir = u;

        lookDir = JoglVectord3.Normal(lookDir);
        udir = JoglVectord3.Normal(udir);

        JoglVectord3 sdir = JoglVectord3.Cross(lookDir, udir);
        JoglVectord3 correctedudir = JoglVectord3.Cross(sdir, lookDir);

        sdir = JoglVectord3.Normal(sdir);
        correctedudir = JoglVectord3.Normal(correctedudir);

        JoglMatrixd4x4 newRotMatrix = new JoglMatrixd4x4();
        
        newRotMatrix.SetValue(0, 0, sdir.getValue(0));
        newRotMatrix.SetValue(0, 1, sdir.getValue(1));
        newRotMatrix.SetValue(0, 2, sdir.getValue(2));
        newRotMatrix.SetValue(0, 3, 0);

        newRotMatrix.SetValue(1, 0, correctedudir.getValue(0));
        newRotMatrix.SetValue(1, 1, correctedudir.getValue(1));
        newRotMatrix.SetValue(1, 2, correctedudir.getValue(2));
        newRotMatrix.SetValue(1, 3, 0);

        newRotMatrix.SetValue(2, 0, -lookDir.getValue(0));
        newRotMatrix.SetValue(2, 1, -lookDir.getValue(1));
        newRotMatrix.SetValue(2, 2, -lookDir.getValue(2));
        newRotMatrix.SetValue(2, 3, 0);

        newRotMatrix.SetValue(3, 0, 0);
        newRotMatrix.SetValue(3, 1, 0);
        newRotMatrix.SetValue(3, 2, 0);
        newRotMatrix.SetValue(3, 3, 1);

        //return JoglMatrixd4x4.CreateTranspose(newRotMatrix);
        return newRotMatrix;
    }
	
	public static JoglMatrixd4x4 CreateTranslation(double x, double y, double z)
	{
		JoglMatrixd4x4 mat = CreateIdentity();

        mat.SetValue(0, 3, x);
        mat.SetValue(1, 3, y);
        mat.SetValue(2, 3, z);
        
        return mat;
	}
	
	public static JoglMatrixd4x4 CreateTranslation(JoglVectord3 t)
	{
		return CreateTranslation(t.getX(), t.getY(), t.getZ());
	}
	
	public static JoglMatrixd4x4 CreateScale(double s)
	{
		return CreateScale(s, s, s);
	}
	
	public static JoglMatrixd4x4 CreateScale(double x, double y, double z)
	{
		JoglMatrixd4x4 mat = CreateIdentity();

        mat.SetValue(0, 0, x);
        mat.SetValue(1, 1, y);
        mat.SetValue(2, 2, z);
        
        return mat;
	}
	
	public static JoglMatrixd4x4 CreatePerspectiveProjection(double fov, double aspect, double nearDist, double farDist)
	{
		//adapted from http://nehe.gamedev.net/article/replacement_for_gluperspective/21002/
		double fH = Math.tan(fov * 0.5 * 0.0174532925) * nearDist;
		double fW = fH * aspect;

		JoglMatrixd4x4 mat = JoglMatrixd4x4.CreateFrustum(nearDist, farDist, -fW, fW, -fH, fH);
		return mat;
	}

	public static JoglMatrixd4x4 CreatePerspectiveProjection(double fovY, double aspect, double zNear, double zFar, double left, double right, double top, double bottom) 
	{
		//adapted from http://nehe.gamedev.net/article/replacement_for_gluperspective/21002/
		double fH = Math.tan(fovY * 0.5 * 0.0174532925) * zNear;
		double fW = fH * aspect;

		double w = fW * 2.0;
		double h = fH * 2.0;
		
		double l = -fW;
		//double r = fW;
		double b = -fH;
		//double t = fH;
		
		JoglMatrixd4x4 mat = JoglMatrixd4x4.CreateFrustum(zNear, zFar, l + (w * left), l + (w * right), b + (h * bottom), b + (h * top));
		return mat;
	}
	
	public static JoglMatrixd4x4 CreateFrustum(double nearDist, double farDist, double l, double r, double b, double t)
	{
	    //
	    // General form of the Projection Matrix
	    //
	    // uh = Cot( fov/2 ) == 1/Tan(fov/2)
	    // uw / uh = 1/aspect
	    // 
	    //   uw         0       0       0
	    //    0        uh       0       0
	    //    0         0      f/(f-n)  1
	    //    0         0    -fn/(f-n)  0
	    //
	    // Make result to be identity first
		
		JoglMatrixd4x4 result = JoglMatrixd4x4.CreateIdentity();

	    result.SetValue(0, 0, (2 * nearDist) / (r - l));
	    result.SetValue(0, 2, (r + l) / (r - l));
	    result.SetValue(1, 1, (2 * nearDist) / (t - b));
	    result.SetValue(1, 2, (t + b) / (t - b));
	    result.SetValue(2, 2, (-(farDist+nearDist)) / (farDist-nearDist));
	    result.SetValue(2, 3, (-2 * farDist * nearDist)/(farDist-nearDist));
	    result.SetValue(3, 2, -1);
	    result.SetValue(3, 3, 0);

        return result;
	}
	
	public JoglVectord3 Transformv3(JoglVectord3 v)
	{
    	JoglVectord3 nv = new JoglVectord3();
    	double vx = v.getValue(0);
    	double vy = v.getValue(1);
    	double vz = v.getValue(2);
	        
    	nv.setValue(0, (this.GetValue(0,0) * vx) + (this.GetValue(0,1) * vy) + (this.GetValue(0,2) * vz) + (this.GetValue(0,3)));
    	nv.setValue(1, (this.GetValue(1,0) * vx) + (this.GetValue(1,1) * vy) + (this.GetValue(1,2) * vz) + (this.GetValue(1,3)));
    	nv.setValue(2, (this.GetValue(2,0) * vx) + (this.GetValue(2,1) * vy) + (this.GetValue(2,2) * vz) + (this.GetValue(2,3)));
 	
		return nv;
	}
	
	public JoglVectord4 Transformv4(JoglVectord4 v)
	{
    	JoglVectord4 nv = new JoglVectord4();

    	double vx = v.getValue(0);
    	double vy = v.getValue(1);
    	double vz = v.getValue(2);
    	double vu = v.getValue(3);
    	
    	nv.setValue(0, (this.GetValue(0,0) * vx) + (this.GetValue(0,1) * vy) + (this.GetValue(0,2) * vz) + (this.GetValue(0,3) * vu));
    	nv.setValue(1, (this.GetValue(1,0) * vx) + (this.GetValue(1,1) * vy) + (this.GetValue(1,2) * vz) + (this.GetValue(1,3) * vu));
    	nv.setValue(2, (this.GetValue(2,0) * vx) + (this.GetValue(2,1) * vy) + (this.GetValue(2,2) * vz) + (this.GetValue(2,3) * vu));
    	nv.setValue(3, (this.GetValue(3,0) * vx) + (this.GetValue(3,1) * vy) + (this.GetValue(3,2) * vz) + (this.GetValue(3,3) * vu));
	        
		return nv;
	}

	public JoglVectord4 GetRow(int r)
	{
		return new JoglVectord4(GetValue(r, 0), GetValue(r, 1), GetValue(r, 2), GetValue(r, 3));
	}
	
	public void SetRow(JoglVectord4 v, int r)
	{
		SetValue(r, 0, v.getX());
		SetValue(r, 1, v.getY());
		SetValue(r, 2, v.getZ());
		SetValue(r, 3, v.getW());
	}
	
    public static JoglMatrixd4x4 Inverse(JoglMatrixd4x4 elements) throws Exception
    {
    	JoglMatrixd4x4 nm;
    	JoglMatrixd4x4 tmp;

        nm = new JoglMatrixd4x4();
        tmp = JoglMatrixd4x4.CreateIdentity();

        for (int i = 0; i < 4; i++)
        {
            for (int j = 0; j < 4; j++)
            {
                nm.SetValue(i, j, elements.GetValue(i, j));
            }
        }

        //Gaussian Elemination
        //========================

        //matrix to upper triangular form
        for (int k = 0; k < 4; k++)
        {
            //elemination
            for (int i = k + 1; i < 4; i++)
            {
                if (i > k && nm.GetValue(i, k) != 0)
                {
                    double q = nm.GetValue(i, k) / nm.GetValue(k, k);
                    JoglVectord4 x = nm.GetRow(i);
                    JoglVectord4 y = nm.GetRow(k);

                    //nm.SetRow(x - (y * q), i);
                    nm.SetRow(JoglVectord4.Subtract(x, JoglVectord4.Multiply(y , q)), i);

                    JoglVectord4 z = tmp.GetRow(i);
                    JoglVectord4 w = tmp.GetRow(k);

                    //tmp.SetRow(z - (w * q), i);
                    tmp.SetRow(JoglVectord4.Subtract(z, JoglVectord4.Multiply(w , q)), i);
                }
            }
        }
        
        //find determinant
        double det = 1;
        for (int i = 0; i < 4; i++)
        {
            det *= nm.GetValue(i, i);
        }

        if (det == 0)
            throw new Exception("Matrix has no inverse: Determinant equals zero!");

        //normalization
        for (int i = 0; i < 4; i++)
        {
            if (nm.GetValue(i, i) != 1)
            {
                double q = nm.GetValue(i, i);

                JoglVectord4 x = nm.GetRow(i);
                //nm.SetRow(x / q, i);
                nm.SetRow(JoglVectord4.Divide(x, q), i);

                JoglVectord4 y = tmp.GetRow(i);
                //tmp.SetRow(y / q, i);
                tmp.SetRow(JoglVectord4.Divide(y, q), i);
            }
        }

        //matrix to identity
        for (int k = 1; k < 4; k++)
        {
            //elemination
            for (int i = 0; i < 4; i++)
            {
                if (i < k && nm.GetValue(i, k) != 0)
                {
                    double q = nm.GetValue(i, k) / nm.GetValue(k, k);
                    JoglVectord4 x = nm.GetRow(i);
                    JoglVectord4 y = nm.GetRow(k);

                    //nm.SetRow(x - (y * q), i);
                    nm.SetRow(JoglVectord4.Subtract(x, JoglVectord4.Multiply(y , q)), i);

                    JoglVectord4 z = tmp.GetRow(i);
                    JoglVectord4 w = tmp.GetRow(k);

                    //tmp.SetRow(z - (w * q), i);
                    tmp.SetRow(JoglVectord4.Subtract(z, JoglVectord4.Multiply(w , q)), i);
                }
            }
        }
        return tmp;
    }
    
    public static String print(JoglMatrixd4x4 mat)
    {
    	String s = "";
    	for (int i = 0; i < 16; i++)
    	{
    		if (i > 0)
    		{
    			s += ", ";
    		}
    		s += mat.data[i];
    	}
    	
    	return "[" + s + "]";
    }
}
