package org.processmining.visualisation3d.settings;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.prefs.Preferences;

import org.processmining.visualisation3d.GraphicsPipeline;
import org.processmining.visualisation3d.gldatastructures.JoglMultiSampleFrameBufferObject;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2;

public class JoglGraphicsSettings 
{
	private int multiSamples = 4;
	private int maxMultiSamples = 1;
	private int anisoSamples = 16;
	private int maxAnisoSamples = 1;
	public boolean bHQShaders = true;
	public boolean bHQGeometry = true;
	
//	public JoglGraphicsSettings()
//	{
//	}
	
	public JoglGraphicsSettings(GL2 gl)
	{
		maxMultiSamples = JoglMultiSampleFrameBufferObject.GetMaxSamples(gl);
		maxAnisoSamples = GetMaxAnisotropicSamples(gl);
		
		setMultiSamples(multiSamples);
		setAnisoSamples(anisoSamples);
	}
	
	public void setMultiSamples(int sNum)
	{
		multiSamples = Math.min(sNum, maxMultiSamples);
	}
	
	public void setAnisoSamples(int sNum)
	{
		anisoSamples = Math.min(sNum, maxAnisoSamples);
	}
	
	public static JoglGraphicsSettings GetDefaultSettings(GL2 gl)
	{
		JoglGraphicsSettings mySettings = null;
		String vendor = gl.glGetString(GL.GL_VENDOR);
		String gpuName = gl.glGetString(GL.GL_RENDERER);
		
		if (vendor.startsWith("Intel"))
		{
			System.out.println("Auto-detection: Using Low graphics settings.");
			mySettings = GetLowSettings(gl);
		}
		else if (vendor.startsWith("NVIDIA") && gpuName.contains("GTX 980"))
		{
			System.out.println("Auto-detection: Using Max graphics settings.");
			mySettings = GetMaxSettings(gl);
		}
		else
		{
			System.out.println("Auto-detection: Using Default graphics settings.");
			mySettings = new JoglGraphicsSettings(gl);
		}
		
		//try loading default settings
		mySettings.LoadPreferences();
		
		return mySettings;
	}
	
	public void LoadPreferences()
	{
		Preferences prefs = Preferences.userNodeForPackage(this.getClass());
		
		setMultiSamples(prefs.getInt("MultiSamples", multiSamples));
		setAnisoSamples(prefs.getInt("AnisoSamples", anisoSamples));
		bHQGeometry = prefs.getBoolean("HQGeometry", bHQGeometry);
		bHQShaders = prefs.getBoolean("HQShaders", bHQShaders);
	}
	
	public void SavePreferences()
	{
		Preferences prefs = Preferences.userNodeForPackage(this.getClass());
		
		prefs.putInt("MultiSamples", multiSamples);
		prefs.putInt("AnisoSamples", anisoSamples);
		prefs.putBoolean("HQGeometry", bHQGeometry);
		prefs.putBoolean("HQShaders", bHQShaders);
	}
	
	public static JoglGraphicsSettings GetMaxSettings(GL2 gl)
	{
		JoglGraphicsSettings jgs = new JoglGraphicsSettings(gl);
		jgs.setMultiSamples(JoglMultiSampleFrameBufferObject.GetMaxSamples(gl));
		jgs.setAnisoSamples(GetMaxAnisotropicSamples(gl));
		return jgs;
	}
	
	public static JoglGraphicsSettings GetLowSettings(GL2 gl)
	{
		JoglGraphicsSettings jgs = new JoglGraphicsSettings(gl);
		jgs.setMultiSamples(1);
		jgs.setAnisoSamples(1);
		jgs.bHQGeometry = false;
		jgs.bHQShaders = false;
		return jgs;
	}
	
	public static boolean SupportsExtension(GraphicsPipeline pipe, String ext)
	{
		return SupportsExtension(pipe.gl, ext);
	}
	
	private static boolean SupportsExtension(GL2 gl, String ext)
	{
		String str = gl.glGetString(GL.GL_EXTENSIONS);
		
		int idx = str.indexOf(ext);
		return (idx >= 0);
	}
	
	public static boolean SupportsVertexbufferObjects(GraphicsPipeline pipe)
	{
		return SupportsVertexbufferObjects(pipe.gl);
	}
	
	private static boolean SupportsVertexbufferObjects(GL2 gl)
	{
		if (!JoglGraphicsSettings.SupportsExtension(gl, "GL_ARB_vertex_buffer_object"))
		{
			return false;
		}

		if (!JoglGraphicsSettings.SupportsExtension(gl, "GL_ARB_vertex_array_object"))
		{
			return false;
		}
		
		try
		{
			IntBuffer rbtest = IntBuffer.allocate(1);
			gl.glGenBuffers(1, rbtest);
			gl.glDeleteBuffers(1, rbtest);
			return true;
		}
		catch (Exception ex)
		{
			
		}
		return false;
	}
	
	public static boolean SupportsFramebufferObjects(GraphicsPipeline pipe)
	{
		return SupportsFramebufferObjects(pipe.gl);
		//return JoglGraphicsSettings.SupportsExtension(gl, "GL_ARB_framebuffer_object");
	}
	
	private static boolean SupportsFramebufferObjects(GL2 gl)
	{
		return JoglGraphicsSettings.SupportsExtension(gl, "GL_ARB_framebuffer_object");
	}
	
	public static boolean SupportsShaderObjects(GraphicsPipeline pipe)
	{
		return SupportsShaderObjects(pipe.gl);
		//return (JoglGraphicsSettings.SupportsExtension(gl, "ARB_shader_objects") && JoglGraphicsSettings.SupportsExtension(gl, "ARB_vertex_shader") && JoglGraphicsSettings.SupportsExtension(gl, "ARB_fragment_shader"));
	}
	
	private static boolean SupportsShaderObjects(GL2 gl)
	{
		return (JoglGraphicsSettings.SupportsExtension(gl, "ARB_shader_objects") && JoglGraphicsSettings.SupportsExtension(gl, "ARB_vertex_shader") && JoglGraphicsSettings.SupportsExtension(gl, "ARB_fragment_shader"));
	}
	
	public static boolean SupportsAnisotropicFiltering(GraphicsPipeline pipe)
	{
		return SupportsAnisotropicFiltering(pipe.gl);
		//return (JoglGraphicsSettings.SupportsExtension(gl, "GL_EXT_texture_filter_anisotropic"));
	}
	
	private static boolean SupportsAnisotropicFiltering(GL2 gl)
	{
		return (JoglGraphicsSettings.SupportsExtension(gl, "GL_EXT_texture_filter_anisotropic"));
	}
	
	public int GetAntiAliasingSamples()
	{
        return multiSamples;
	}
	
	public int GetAnisotropicSamples()
	{
        return anisoSamples;
	}
	
	public static int GetMaxAnisotropicSamples(GraphicsPipeline pipe)
	{
		return GetMaxAnisotropicSamples(pipe.gl);
	}
	
	private static int GetMaxAnisotropicSamples(GL2 gl)
	{
		FloatBuffer fb = FloatBuffer.allocate(1);
        gl.glGetFloatv(GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, fb);
        return (int)fb.get(0);
	}
	
	public static void CheckPlatformSupport(GL2 gl)
	{
		System.out.println("Graphics platform: " + gl.glGetString(GL.GL_VENDOR) + ", " + gl.glGetString(GL.GL_RENDERER));
		
		if (SupportsVertexbufferObjects(gl))
		{
			System.out.println("Platform supports vertex buffer objects.");
		}
		else
		{
			System.out.println("Platform does not support vertex buffer objects.");
		}
		
		if (SupportsFramebufferObjects(gl))
		{
			System.out.println("Platform supports frame buffer objects.");
		}
		else
		{
			System.out.println("Platform does not support frame buffer objects.");
		}

		if (SupportsShaderObjects(gl))
		{
			System.out.println("Platform supports shader objects.");
		}
		else
		{
			System.out.println("Platform does not support shader objects.");
		}

		if (SupportsAnisotropicFiltering(gl))
		{
			System.out.println("Platform supports anisotropic filtering.");
	        System.out.println("Max anisotropic samples supported: " + GetMaxAnisotropicSamples(gl));
		}
		else
		{
			System.out.println("Platform does not support anisotropic filtering.");
		}

		
		int maxSamples = JoglMultiSampleFrameBufferObject.GetMaxSamples(gl);
		System.out.println("Max samples supported: " + maxSamples);
	}

}
